/*
 * Copyright (c) 1990 AT&T All Rights Reserved
 *	This is unpublished proprietary source code of AT&T
 *	The copyright notice above does not evidence any
 *	actual or intended publication of such source code.
 *
 * Pi Unix driver for Silicon Graphics 4-D Series of Workstations
 *	This file uses /debug to examine running processes.
 *
 * D. A. Kapilow	1/20/92
 */
#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include <filehdr.h>
#include <scnhdr.h>
#include <core.out.h>
#include <time.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/procfs.h>
#include <sys/wait.h>	/* For wait */
#include <sys/param.h>
#include "dbmonitor/dbhost.h"

#define	bit(x)		(1<<(x-1))
#define	MIPS_SYSCALL	0x0000000C
#define	MIPS_BPT	0x0000000D
#define REGSIZE		(39 * 4)
#define	REG_PC		32
#define	ISCORE(p)	((p)->pid < -1)

/*
 * Data structure private to each process or core dump
 */
struct coreseg {
	int fd;
	unsigned long	start;
	unsigned long	end;
	unsigned long	offset;
};

struct localproc {
	int		pid;
	int		corefd;
	int		opencnt;
	struct	Unixstate	state;
	struct localproc	*next;
	prstatus_t	s;
	/* For core dumps */
	int		stabfd;		/* text */
	int		nseg;
	struct coreseg	*seg;
};

/*
 * Register layout transmitted to pi
 */
struct dbregs {
	int r[32];
	int pc;
	int cause;
	int badvaddr;
	int mmhi;
	int mmlo;
	int fpc_csr;
	int fpc_eir;
};

struct localproc *phead;
static char procname[] = "/debug/XXXXX";
static char procbuffer[128];
static sysset_t entrymask;
static int hangpid;
long pi_corepid = -1;
static unsigned long regaddr = -REGSIZE - 4;
static long bkpt_instr = MIPS_BPT;
static char *pscmds[] = {
	"/bin/ps -u XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
	"/bin/ps -a    ",
	"/bin/ps -e    ",
	0
};

/* Routines local to this file */
static void state();
static int waithandler();
static char *regrw();
static void coreclose();
static char *corerw();

/* Also use for initialization */
char *pi_osname()
{
	struct passwd *pw;
	static int first = 1;

	if (first) {
		close(2); /* Otherwise error messages from ps goto rtpi */
		open("/dev/null", 2);
		pw = getpwuid(getuid());
		if (pw)
			sprintf(pscmds[0], "/bin/ps -u %s", pw->pw_name);
		else
			sprintf(pscmds[0], "/bin/ps");
		signal(SIGCLD, waithandler);
		first = 0;
	}
	return "IRIX Release 4.0.5 System V";
}

char	**pi_getpscmds()	{ return pscmds; }
int	pi_getpsfield()		{ return 15; }
long	pi_regaddr()		{ return (long)regaddr; }
long	pi_scratchaddr()	{ return 0; }	/* FIX ME */
int	pi_nsig()		{ return NUMSIGS; }
int	pi_exechangsupported()	{ return 1; }
int	pi_stabfdsupported()	{ return 1; }

long pi_sigmaskinit()
{
	return	bit(SIGILL)|bit(SIGINT)|bit(SIGTRAP)|bit(SIGIOT)|
		bit(SIGEMT)|bit(SIGFPE)|bit(SIGBUS)|bit(SIGSEGV)|
		bit(SIGSYS)|bit(SIGPIPE)|bit(SIGSTOP);
}

struct localproc *pi_pidtoproc(pid)
{
	register struct localproc *p;

	for (p = phead; p; p = p->next)
		if (p->pid == pid)
			return p;
	return 0;
}

struct localproc *pi_open(pid)
{
	register struct localproc *p = phead;
	sysset_t sysmask;
	int wasactive = 0;

	if (pid < 0)
		return 0;
	if (p = pi_pidtoproc(pid)) {
		p->opencnt++;
		return p;
	}
	p = (struct localproc *)malloc(sizeof(struct localproc));
	if (!p)
		return 0;
	sprintf(procname,"/debug/%05d", pid);
	if ((p->corefd = open(procname, 2)) < 0) {
		free(p);
		return 0;
	}
	p->pid = pid;
	p->opencnt = 1;
	p->next = phead;
	phead = p;
	state(p);
	premptyset(&sysmask);
	ioctl(p->corefd, PIOCSENTRY, &sysmask);
	ioctl(p->corefd, PIOCSEXIT, &sysmask);
	return p;
}

void pi_close(p)
struct localproc *p;
{
	register struct localproc *q;
	long mode;
	sysset_t sysmask;
	sigset_t sset;

	if (--p->opencnt > 0)
		return;
	if (ISCORE(p))
		coreclose(p);
	else {
		if (p->state.state != UNIX_ERRORED) {
			premptyset(&sysmask);
			ioctl(p->corefd, PIOCSENTRY, &sysmask);
			ioctl(p->corefd, PIOCSEXIT, &sysmask);
			premptyset(&sset);
			ioctl(p->corefd, PIOCSTRACE, &sset);
			mode = PR_RLC | PR_FORK;
			ioctl(p->corefd, PIOCRESET, &mode);
		}
		close(p->corefd);
	}
	if (p == phead)
		phead = p->next;
	else {
		for (q = phead; q->next != p; q = q->next)
			;
		q->next = p->next;
	}
	free(p);
}

int pi_getsymtabfd(p)
struct localproc *p;
{
	return ioctl(p->corefd, PIOCOPENM, 0);
}

char *pi_run(p)
struct localproc *p;
{
	ioctl(p->corefd, PIOCRUN, 0);
	return 0;
}

char *pi_stop(p)
struct localproc *p;
{
	ioctl(p->corefd, PIOCSTOP, 0);
	return 0;
}

char *pi_destroy(p)
struct localproc *p;
{
	int sig = SIGKILL;
	ioctl(p->corefd, PIOCKILL, &sig);
	state(p);
	p->state.state = UNIX_ERRORED;
	p->state.code = SIGKILL;
	return 0;
}

void pi_getstate(p, s)
struct localproc *p;
struct	Unixstate *s;
{
	if (!ISCORE(p))
		state(p);
	*s = p->state;
}

static void state(p)
struct localproc *p;
{
	if (ioctl(p->corefd, PIOCSTATUS, &p->s)) {
		/*
		 * If the call failed, it is probably because the process has
		 * disappeared. With the SGI /debug, there is no way to
		 * determine the exist status or reason the program died, so
		 * set it to make it look like a signal that cannot really
		 * happen was sent. If the server is the parent of the process,
		 * handling the SIGCLD signal will set the correct state value.
		 * Otherwise, the setting below remains.
		 */
		if (p->state.state != UNIX_ERRORED) {
			p->state.state = UNIX_ERRORED;
			p->state.code = NSIG;
		}
		return;
	}
	if (!(p->s.pr_flags & PR_STOPPED))
		p->state.state = UNIX_ACTIVE;
	else if (p->s.pr_why == PR_SIGNALLED) {
		if (p->s.pr_what == SIGTRAP) {
			p->state.state = UNIX_BREAKED;
			ioctl(p->corefd, PIOCSSIG, 0);
		} else {
			p->state.state = UNIX_PENDING;
			p->state.code = p->s.pr_what;
		}
	} else
		p->state.state = UNIX_HALTED;
}

char *pi_readwrite(p, buf, addr, r, w)
struct localproc *p;
char *buf;
unsigned long addr;
{
	int count, ret;

	if (ISCORE(p))
		return corerw(p, buf, addr, r, w);
	if (addr >= regaddr)
		return regrw(p, buf, (int)(addr - regaddr), r, w);
	if (lseek(p->corefd, addr, 0) == -1)
		return "lseek on /debug failed";
	if (w) {
		ret = write(p->corefd, buf, w);
		count = w;
	} else {
		ret = read(p->corefd, buf, r);
		count = r;
	}
	if (ret != count)
		 return "read/write on /debug failed";
	return 0;
}

static char *regrw(p, buf, offset, r, w)
struct localproc *p;
char *buf;
{
	struct dbregs reg;
	gregset_t procreg;
	char *error = 0;
	int wasactive = 0;

	if (p->state.state == UNIX_ACTIVE) {
		ioctl(p->corefd, PIOCSTOP, 0);
		state(p);
		if (p->s.pr_why == PR_REQUESTED)
			wasactive = 1;
	}
	if (ioctl(p->corefd, PIOCGREG, &procreg)) {
		error =  "DFCGETREGS failed";
		goto out;
	}
	bcopy((char*)procreg.gp_regs, (char*)reg.r, sizeof(reg.r));
	reg.pc = procreg.gp_pc;
	reg.cause = procreg.gp_cause;
	reg.mmhi = procreg.gp_mdhi;
	reg.mmlo = procreg.gp_mdlo;
	reg.cause = reg.fpc_csr = reg.fpc_eir = 0;
	if (w) {
		bcopy(buf, (char*)&reg + offset, w);
		bcopy((char*)reg.r, (char*)procreg.gp_regs, sizeof(reg.r));
		procreg.gp_pc = reg.pc;
		procreg.gp_cause = reg.cause;
		procreg.gp_mdhi = reg.mmhi;
		procreg.gp_mdlo = reg.mmlo;
		ioctl(p->corefd, PIOCSREG, &procreg);
	} else
		bcopy((char*)&reg + offset, buf, r);
out:
	if (wasactive)
		ioctl(p->corefd, PIOCRUN, 0);
	return error;
}

char *pi_setbkpt(p, addr)
struct localproc *p;
long addr;
{
	return pi_readwrite(p,(char*)&bkpt_instr, addr, 0, sizeof(bkpt_instr));
}

char *pi_step(p)
struct localproc *p;
{
	prrun_t prrun;

	prrun.pr_flags = PRSTEP;
	if (ioctl(p->corefd, PIOCRUN, &prrun))
		return "DFCSSTEP error";
	if (ioctl(p->corefd, PIOCWSTOP, 0))
		return "DFCWSTOP error";
	return 0;
}

static void alarmcatcher()	{}

char *pi_waitstop(p)
struct localproc *p;
{
	char *err = 0;
	void (*save)();

	save = signal(SIGALRM, alarmcatcher);
	alarm(15);
	if (ioctl(p->corefd, PIOCWSTOP, 0))
		err = "timeout waiting for breakpoint";
	alarm(0);
	signal(SIGALRM, save);
	return err;
}

char *pi_ssig(p, sig)
struct localproc *p;
long sig;
{
	ioctl(p->corefd, PIOCKILL, &sig);
	return 0;
}

char *pi_csig(p)
struct localproc *p;
{
	ioctl(p->corefd, PIOCSSIG, 0);
	return 0;
}

char *pi_sigmask(p, mask)
struct localproc *p;
long mask;
{
	sigset_t sset;
	int i;

	premptyset(&sset);
	for(i = 1; i <= 32; i++)
		if (mask & bit(i))
			praddset(&sset, i);
	ioctl(p->corefd, PIOCSTRACE, &sset);
	return 0;
}

char *pi_signalname(sig)
long sig;
{
	char *cp;

	switch (sig) {
	case SIGHUP:	cp = "hangup"; break;
	case SIGINT:	cp = "interrupt"; break;
	case SIGQUIT:	cp = "quit"; break;
	case SIGILL:	cp = "illegal instruction"; break;
	case SIGTRAP:	cp = "trace trap"; break;
	case SIGABRT:	cp = "abort"; break;
	case SIGEMT:	cp = "EMT instruction"; break;
	case SIGFPE:	cp = "floating exception"; break;
	case SIGKILL:	cp = "kill"; break;
	case SIGBUS:	cp = "bus error"; break;
	case SIGSEGV:	cp = "segmentation violation"; break;
	case SIGSYS:	cp = "bad system call"; break;
	case SIGPIPE:	cp = "broken pipe"; break;
	case SIGALRM:	cp = "alarm call"; break;
	case SIGTERM:	cp = "terminated"; break;
	case SIGUSR1:	cp = "user signal 1"; break;
	case SIGUSR2:	cp = "user signal 2"; break;
	case SIGCHLD:	cp = "child termination"; break;
	case SIGPWR:	cp = "power-fail"; break;
	case SIGSTOP:	cp = "stop"; break;
	case SIGTSTP:	cp = "tty stop"; break;
	case SIGPOLL:	cp = "pollable event"; break;
	case SIGIO:	cp = "i/o possible"; break;
	case SIGURG:	cp = "urgent i/o condition"; break;
	case SIGWINCH:	cp = "window changed"; break;
	case SIGVTALRM:	cp = "virtual time alarm"; break;
	case SIGPROF:	cp = "profiling alarm"; break;
	case SIGCONT:	cp = "continue"; break;
	case SIGTTIN:	cp = "stop tty input"; break;
	case SIGTTOU:	cp = "stop tty output"; break;
	case SIGXCPU:	cp = "cpu time limit"; break;
	case SIGXFSZ:	cp = "file size limit"; break;
	case NSIG:	cp = "cannot determine cause"; break;
	default:
		cp = procbuffer;
		sprintf(procbuffer, "signal %d", sig);
		break;
	}
	return cp;
}

char *pi_proctime(p)
struct localproc *p;
{
	sprintf(procbuffer, "%d.%02du %d.%02ds",
		p->s.pr_utime.tv_sec, p->s.pr_utime.tv_nsec/10000000,
		p->s.pr_stime.tv_sec, p->s.pr_stime.tv_nsec/10000000);
	return procbuffer;
}

int pi_getppid(p)
struct localproc *p;
{
	return (int)p->s.pr_ppid;
}

int pi_atsyscall(p)
struct localproc *p;
{
	long pc, instr; 

	pi_readwrite(p, (char *)&pc, regaddr + REG_PC * 4, sizeof(pc), 0);
	pi_readwrite(p, (char *)&instr, pc, sizeof(instr), 0);
	return (instr == MIPS_SYSCALL);
}

char *pi_exechang(p, e)
struct localproc *p;
long e;
{
	sysset_t sysmask;

	ioctl(p->corefd, PIOCGEXIT, &sysmask);
	if (e) {
		praddset(&sysmask, SYS_execv);
		praddset(&sysmask, SYS_execve);
	}
	else {
		prdelset(&sysmask, SYS_execv);
		prdelset(&sysmask, SYS_execve);
	}
	ioctl(p->corefd, PIOCSEXIT, &sysmask);
	return 0;
}

/*
 * Expand a path - this function is needed because in release
 * 3.3.2, if a process sets DFCSEXEC and calls execvp, the process
 * will hang even if the exec call fails. For some reason this leads
 * to runaway core dumps. This is a kludge to locate the file first,
 * to make sure it is executable and readable.  This way pi_hang can
 * call execv instead of the nasty execvp. Remove this routine when
 * /debug has saner behavior.
 */
static char *pathexpand(f, path, a)
char *f, *path;
int a;
{
	static char file[128];
	register char *p;

	if (*f != '/' && strncmp(f, "./", 2) && strncmp(f, "../", 3) && 
	    path!=0){
		while(*path){
			for(p=file; *path && *path!=':';)
				*p++ = *path++;
			if(p!=file)
				*p++='/';
			if(*path)
				path++;
			(void)strcpy(p, f);
			if (access(file, a) != -1)
				return file;
		}
	}
	if (access(f, a) != -1 )
		return f;
	return 0;
}

/*
 * This function and the function below are for starting new
 * processes from pi.
 */
int pi_hang(cmd)
char *cmd;
{
	char *argv[10], *cp;
	int i, fd;
	char *file, *getenv();
	struct localproc *p;
	sysset_t sysmask;
	long mode;
	
	i = strlen(cmd);
	if (++i > sizeof(procbuffer)) {
		i = sizeof(procbuffer) - 1;
		procbuffer[i] = 0;
	}
	bcopy(cmd, procbuffer, i);
	argv[0] = cp = procbuffer;
	for(i = 1;;) {
		while(*cp && *cp != ' ')
			cp++;
		if (!*cp) {
			argv[i] = 0;
			break;
		} else {
			*cp++ = 0;
			while (*cp == ' ')
				cp++;
			if (*cp)
				argv[i++] = cp;
		}
	}
	if (!(file = pathexpand(argv[0], getenv("PATH"), 5)))
		return 0;
	hangpid = fork();
	if (!hangpid){
		for (fd = 0; fd < NOFILES_MAX; ++fd)
			close(fd);
		open("/dev/null", 2);
		dup2(0, 1);
		dup2(0, 2);
		setpgrp(0, getpid());
		sprintf(procname, "/debug/%05d", getpid());
		if ((fd = open(procname, 2)) < 0)
			exit(1);
		mode = PR_RLC;
		ioctl(fd, PIOCRESET, &mode);
		premptyset(&sysmask);
		praddset(&sysmask, SYS_execv);
		praddset(&sysmask, SYS_execve);
		ioctl(fd, PIOCSEXIT, &sysmask);
		close(fd);
		execv(file, argv);
		exit(0);
	}
	if (hangpid < 0)
		return 0;
	sprintf(procname,"/debug/%05d", hangpid);
	if ((fd = open(procname, 2)) < 0) {
		kill(9, hangpid);
		return 0;
	}
	mode = PR_RLC;
	ioctl(fd, PIOCRESET, &mode);
	p = (struct localproc *)malloc(sizeof(struct localproc));
	if (!p) {
		kill(9, hangpid);
		close(fd);
		return 0;
	}
	p->corefd = fd;
	p->pid = hangpid;
	p->opencnt = 0;
	p->next = phead;
	phead = p;
	ioctl(p->corefd, PIOCWSTOP, 0);
	state(p);
	if (p->s.pr_errno) { /* The exec failed - get rid of it */
		pi_ssig(p, 9);
		state(p);
		pi_close(p);
		return 0;
	}
	premptyset(&sysmask);
	ioctl(p->corefd, PIOCSENTRY, &sysmask);
	ioctl(p->corefd, PIOCSEXIT, &sysmask);
	return hangpid;
}

static int waithandler()
{
	int pid, cursig;
	union wait tstat;
	struct localproc *p;

	pid = wait(&tstat);
	signal(SIGCLD, waithandler);
	if (pid < 0 || !(p = pi_pidtoproc(pid)))
		return;
	if (WIFSTOPPED(tstat)) {
		cursig = tstat.w_stopsig;
		if (cursig == SIGSTOP)
			p->state.state = UNIX_HALTED;
		else if (cursig == SIGTRAP)
			p->state.state = UNIX_BREAKED;
		else {
			p->state.state = UNIX_PENDING;
			p->state.code = cursig;
		}
	} else {
		p->state.state = UNIX_ERRORED;
		p->state.code = tstat.w_status & 0xFFFF;
	}
}

/*
 * The functions below are for debugging core dumps
 */
struct localproc *pi_coreopen(corep, symtabp)
char *corep, *symtabp;
{
	register struct localproc *p;
	register struct coreseg *cs;
	int symtabfd, corefd = -1;
	int mode;
	struct coreout c;
	struct vmap *v;
	int i;
	struct filehdr fh;
	struct scnhdr sh;

	if ((symtabfd = open(symtabp, 0)) < 0)
		return 0;
	for (mode = 2; corefd < 0 && mode >= 0; mode--)
		corefd = open(corep, mode);
	if (corefd < 0 ||
	    read(corefd, (char*)&c, sizeof(c)) != sizeof(c) ||
	    c.c_magic != CORE_MAGIC ||
	    read(symtabfd, (char*)&fh, sizeof(fh)) != sizeof(fh) ||
	    lseek(symtabfd, (long)fh.f_opthdr, 1) == -1 ||
	    read(symtabfd, (char*)&sh, sizeof(sh)) != sizeof(sh) ||
	    strcmp(sh.s_name, ".text")) {
		close(symtabfd);
		return 0;
	}
	p = (struct localproc *)malloc(sizeof(struct localproc));
	v = (struct vmap *)malloc(sizeof(struct vmap) * c.c_nvmap);
	if (!p || !v) {
		if (p)
			free(p);
		close(corefd);
		close(symtabfd);
		return 0;
	}
	p->pid = --pi_corepid;
	p->opencnt = 1;
	p->next = phead;
	phead = p;
	p->corefd = corefd;
	p->stabfd = symtabfd;

        lseek(corefd, c.c_vmapoffset, 0);
        read(corefd, (char*)v, sizeof(struct vmap) * c.c_nvmap);
	for(p->nseg = 0, i = 0; i < c.c_nvmap; i++)
		if (v[i].v_flags & VDUMPED)
			p->nseg++;
	/* Add segments for text, + 2 for registers */
	p->nseg += 3;
	p->seg = cs = (struct coreseg *)
			malloc(sizeof(struct coreseg) * p->nseg);
	if (!cs) {
		free(p);
		free(v);
		close(corefd);
		close(symtabfd);
		return 0;
	}
	/* General purpose registers */
	cs->start = regaddr;
	cs->end = cs->start + sizeof(long) * 32;
	cs->offset = c.c_idesc[I_GPREGS].i_offset;
	cs->fd = corefd;
	cs++;
	/* Special registers */
	cs->start = regaddr + sizeof(long) * 32;
	cs->end = regaddr+REGSIZE;
	cs->offset = c.c_idesc[I_SPECREGS].i_offset;
	cs->fd = corefd;
	cs++;
	/* Text segment */
	cs->start = sh.s_paddr;
	cs->end = cs->start + sh.s_size;
	cs->offset = sh.s_scnptr;
	cs->fd = symtabfd;
	cs++;
	/* Other segments (data, stack, sharedmem) */
	for(i = 0; i < c.c_nvmap; i++) {
		if (v[i].v_flags & VDUMPED) {
			cs->start = v[i].v_vaddr;
			cs->end = cs->start + v[i].v_len;
			cs->offset = v[i].v_offset;
			cs->fd = corefd;
			cs++;
		}
	}
	free(v);
	p->state.state = UNIX_PENDING;
	p->state.code = c.c_sigcause;
	return p;
}

static void coreclose(p)
struct localproc *p;
{
	close(p->corefd);
	close(p->stabfd);
	free(p->seg);
}

static char *corerw(p, buf, addr, r, w)
struct localproc *p;
char *buf;
unsigned long addr;
{
	register struct coreseg *cs, *csend;

	cs = p->seg;
	csend = &cs[p->nseg];
	for (; cs < csend; cs++)
		if (cs->start <= addr && addr < cs->end)
			break;
	if (cs == csend)
		return "corerw:invalid address";
	addr = addr - cs->start + cs->offset;
	if (lseek(cs->fd, addr, 0) == -1)
		return "corerw:lseek failed";
	if (w) {
		if (write(cs->fd, buf, w) != w)
			return "corerw:write error";
	} else {
		if (read(cs->fd, buf, r) != r)
			return "corerw:read error";
	}
	return 0;
}
